"""
A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

from os import path

import codecs
from setuptools import setup, find_packages

HERE = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with codecs.open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()

setup(
    name='anyrl',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.11.6',

    description='A library for Reinforcement Learning',
    long_description=LONG_DESCRIPTION,

    # The project's main homepage.
    url='https://github.com/unixpickle/anyrl-py',

    # Author details
    author='Alex Nichol',
    author_email='unixpickle@gmail.com',

    # Choose your license
    license='BSD',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: BSD License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    # What does your project relate to?
    keywords='ai reinforcement learning',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['examples']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'numpy>=1.0.0,<2.0.0',
        'cloudpickle>=0.4.0',
        'gym>=0.9.2,<0.10.0',
        'pandas>=0.20.0,<0.21.0'
    ],

    extras_require={
        "tf": ["tensorflow>=1.0.0"],
        "tf_gpu": ["tensorflow-gpu>=1.0.0"],
        "mpi": ["mpi4py>=2.0.0,<3.0.0"]
    }
)



"""
Advantage estimation routines.
"""

from abc import ABC, abstractmethod

class AdvantageEstimator(ABC):
    """
    A technique for using a value function to estimate
    the advantage function.
    """
    @abstractmethod
    def advantages(self, rollouts):
        """
        Compute the advantages for the rollouts and return
        a list of advantage lists of the same shape as the
        rollouts.
        """
        pass

    def targets(self, rollouts):
        """
        Compute new targets for the value function.

        The result is the same shape as advantages().
        """
        res = [x.copy() for x in self.advantages(rollouts)]
        for rollout_idx, rollout in enumerate(rollouts):
            for timestep, model_out in enumerate(rollout.step_model_outs):
                value_out = model_out['values'][0]
                res[rollout_idx][timestep] += value_out
        return res

class GAE(AdvantageEstimator):
    """
    An implementation of Generalized Advantage Estimation.
    """
    def __init__(self, lam, discount, target_lam=None):
        self.lam = lam
        self.discount = discount
        self.target_lam = target_lam

    def advantages(self, rollouts):
        res = []
        for rollout in rollouts:
            adv = 0
            advs = []
            for i in range(rollout.num_steps)[::-1]:
                delta = rollout.rewards[i] - rollout.predicted_value(i)
                if i+1 < len(rollout.model_outs):
                    delta += self.discount * rollout.predicted_value(i + 1)
                adv *= self.lam * self.discount
                adv += delta
                advs.append(adv)
            res.append(advs[::-1])
        return res

    def targets(self, rollouts):
        if self.target_lam is None:
            return super(GAE, self).targets(rollouts)
        proxy = GAE(lam=self.target_lam, discount=self.discount)
        return proxy.targets(rollouts)



"""
Abstractions for RL policies and value functions.
"""

from abc import ABC, abstractmethod, abstractproperty

class Model(ABC):
    """
    An abstract RL policy and (optional) value function.
    """
    @abstractproperty
    def stateful(self):
        """
        Return whether or not the model has states.
        """
        pass

    @abstractmethod
    def start_state(self, batch_size):
        """
        Return a batch of start states.

        State batches are represented as an array of shape
        [batch_size x D], or as a tuple of i elements of
        shapes [batch_size x Di].

        If the Model is not stateful, return None.
        """
        pass

    @abstractmethod
    def step(self, observations, states):
        """
        Apply the model for a single timestep in a batch
        of environments.

        Returns a dict with the following keys:
          'actions': batch of actions, one per env
          'states': new states after the step
          'values': (optional) predicted value function
          'action_params': (optional) parameters that were
            fed into the action distribution.
        """
        pass

class TFActorCritic(Model):
    """
    An actor-critic model which is differentiable via
    TensorFlow.

    Every TFActorCritic has an action distribution and
    observation vectorizer, which can be accessed via
    model.action_dist and model.obs_vectorizer.
    """
    def __init__(self, session, action_dist, obs_vectorizer):
        self.session = session
        self.action_dist = action_dist
        self.obs_vectorizer = obs_vectorizer

    @abstractmethod
    def batch_outputs(self):
        """
        Return three TF tensors: actor_outs, critic_outs,
        mask.

        The mask is a Tensor of 0's and 1's, where 1
        indicates that the sample is valid.

        Both mask and critic_outs should be 1-D.
        The actor_outs shape depends on the shape of
        action distribution parameters.

        These tensors are used in conjunction with the
        feed_dict returned by batches().

        This method may be called multiple times.
        """
        pass

    @abstractmethod
    def batches(self, rollouts, batch_size=None):
        """
        Create an iterator of mini-batches for training
        the actor and the critic.

        Each mini-batch is a dict with these keys:
          'rollout_idxs': rollout index for each sample
          'timestep_idxs': timestep index for each sample
          'feed_dict': inputs that the graph depends on

        There is a one-to-one correspondence between
        samples in the batch and values in the Tensors
        produced by batch_outputs.
        Masked samples should have 0's in rollout_idxs and
        timestep_idxs.

        Args:
          rollouts: a list of (partial) rollouts
          batch_size: the approximate mini-batch size
        """
        pass



"""
Conversions to and from Gym spaces.
"""

import gym.spaces as spaces

from .aggregate import TupleDistribution
from .binary import MultiBernoulli
from .categorical import CategoricalSoftmax
from .continuous import BoxGaussian

class UnsupportedGymSpace(Exception):
    """
    Thrown when a Gym space cannot be used as an action
    space.
    """
    def __init__(self, space):
        msg = 'unsupported Gym space: ' + str(space)
        super(UnsupportedGymSpace, self).__init__(msg)
        self.space = space

def gym_space_distribution(space):
    """
    Create a Distribution from a gym.Space.

    If the space is not supported, throws an
    UnsupportedActionSpace exception.
    """
    if isinstance(space, spaces.Discrete):
        return CategoricalSoftmax(space.n)
    elif isinstance(space, spaces.Box):
        return BoxGaussian(space.low, space.high)
    elif isinstance(space, spaces.MultiBinary):
        return MultiBernoulli(space.n)
    elif isinstance(space, spaces.Tuple):
        sub_dists = tuple(gym_space_distribution(s) for s in space.spaces)
        return TupleDistribution(sub_dists)
    elif isinstance(space, spaces.MultiDiscrete):
        discretes = tuple(CategoricalSoftmax(high-low+1, low=low)
                          for low, high in zip(space.low, space.high))
        return TupleDistribution(discretes)
    raise UnsupportedGymSpace(space)

def gym_space_vectorizer(space):
    """
    Create a Vectorizer from a gym.Space.

    If the space is not supported, throws an
    UnsupportedActionSpace exception.
    """
    return gym_space_distribution(space)



"""
Tests for schedules.
"""

# pylint: disable=E1129

import unittest

import numpy as np
import tensorflow as tf

from anyrl.algos import LinearTFSchedule

class TestLinearTFSchedule(unittest.TestCase):
    """
    Tests for LinearTFSchedule.
    """
    def test_values(self):
        """
        Test values throughout the schedule.
        """
        with tf.Graph().as_default():
            with tf.Session() as sess:
                sched = LinearTFSchedule(duration=10, start_value=7, end_value=0.5,
                                         dtype=tf.float64)
                sess.run(tf.global_variables_initializer())
                self.assertTrue(np.allclose(sess.run(sched.value), 7))
                sched.add_time(sess, 7.5)
                self.assertTrue(np.allclose(sess.run(sched.value), 2.125))
                for _ in range(3):
                    sched.add_time(sess, 10)
                    self.assertTrue(np.allclose(sess.run(sched.value), 0.5))

if __name__ == '__main__':
    unittest.main()



"""
Utilities for RL-related tests.
"""

import gym
import gym.spaces as spaces
import numpy as np

from anyrl.models import Model

class SimpleEnv(gym.Env):
    """
    An environment with a pre-determined observation space
    and RNG seed.
    """
    def __init__(self, seed, shape, dtype):
        np.random.seed(seed)
        self._dtype = dtype
        self._start_obs = np.array(np.random.randint(0, 0x100, size=shape),
                                   dtype=dtype)
        self._max_steps = seed + 1
        self._cur_obs = None
        self._cur_step = 0
        self.action_space = gym.spaces.Box(0, 0xff, shape=shape)
        self.observation_space = self.action_space

    def _step(self, action):
        flat_act = np.array(action, dtype=self._dtype)
        self._cur_obs = self._cur_obs + np.roll(flat_act, 1)
        self._cur_step += 1
        done = self._cur_step >= self._max_steps
        reward = self._cur_step / self._max_steps
        return self._cur_obs, reward, done, {'foo': 'bar' + str(reward)}

    def _reset(self):
        self._cur_obs = self._start_obs
        self._cur_step = 0
        return self._cur_obs

class SimpleModel(Model):
    """
    A stateful, deterministic model which is compatible
    with a SimpleEnv.
    """
    def __init__(self, shape, stateful=False, state_tuple=True):
        self.shape = shape
        self._stateful = stateful
        self.state_tuple = state_tuple

    @property
    def stateful(self):
        return self._stateful

    def start_state(self, batch_size):
        if not self.stateful:
            return None
        if self.state_tuple:
            return (np.zeros((batch_size,) + self.shape),
                    np.zeros((batch_size, 1)))
        return np.zeros((batch_size,) + self.shape)

    def step(self, observations, states):
        actions = []
        values = []
        if not self.stateful:
            new_states = None
        elif not self.state_tuple:
            new_states = []
        else:
            new_states = ([], [])
        for i, obs in enumerate(observations):
            if not self.stateful:
                actions.append(obs*3)
            elif self.state_tuple:
                actions.append((states[0][i] + obs) * states[1][i])
                new_states[0].append(states[0][i] + 2*obs)
                new_states[1].append(3 - states[1][i])
            else:
                actions.append(states[i] + obs)
                new_states.append(states[i] + 2*obs)
            values.append(np.sum(obs))
        return {
            'actions': actions,
            'values': values,
            'states': new_states
        }

class TupleCartPole(gym.Env):
    """
    A version of Gym's CartPole-v0 with tuple spaces.

    Intended to stress test models against weird kinds of
    spaces.
    """
    def __init__(self):
        self._inner_env = gym.make('CartPole-v0')
        self.action_space = spaces.Tuple([spaces.Discrete(2), spaces.Discrete(3)])
        obs_space = self._inner_env.observation_space
        self.observation_space = spaces.Tuple([
            spaces.Box(obs_space.low[:3], obs_space.high[:3]),
            spaces.Box(obs_space.low[3:], obs_space.high[3:])
        ])

    def _reset(self):
        return self._split_obs(self._inner_env.reset())

    def _step(self, action):
        obs, rew, done, info = self._inner_env.step(action[0])
        return self._split_obs(obs), rew, done, info

    # pylint: disable=R0201
    def _split_obs(self, obs):
        obs = np.array(obs)
        return (obs[:3], obs[3:])
